/*---------------------------------------------*/
/*	GDP15 - 2014/15
 *
 * 	Author:		Stuart Witts
 * 	Email:		stw1g11@ecs.soton.ac.uk
 * 	Date:		January 2014
 *
 *  -----------------------------------------
 * 	Base Unit embedded software - Header File
 * 	-----------------------------------------
 *
 * */
/*---------------------------------------------*/
  

/*---------------------------------------------*/
/* Define to prevent recursive inclusion */
/* (Idea from STMicroelectronics examples) */
/*---------------------------------------------*/
#ifndef __MAIN_H
#define __MAIN_H


/*---------------------------------------------*/
/* Includes */
/*---------------------------------------------*/

#include "stm32l1xx.h"
#include "stm32l100c_discovery.h"
#include "ff.h"
#include "diskio.h"
#include "stdio.h"
#include "string.h"


/*---------------------------------------------*/
/* Definitions */
/*---------------------------------------------*/

#define DEBUG_USE_USART1
#define SPI2_MASTER			// Sensors
#define SPI3_MASTER			// SD Card

#define TIMEOUT_MS_USART 200
#define TIMEOUT_MS_SPI 157
#define TIMEOUT_MS_SENSOR 800

#define MAX_CHARS 256

#define RTC_DEVICE_ADDR 0xD0
#define RTC_ADDR_SECONDS 0x00
#define RTC_ADDR_MINUTES 0x01
#define RTC_ADDR_HOURS 0x02
#define RTC_ADDR_DAY 0x03
#define RTC_ADDR_DATE 0x04
#define RTC_ADDR_MONTH 0x05
#define RTC_ADDR_YEARS 0x06
#define RTC_ADDR_CFG1 0x07
#define RTC_ADDR_TCH2 0x08
#define RTC_ADDR_CFG2 0x09

// SENSOR_MAX_BYTES = maximum data bytes for any type of sensor
#define SENSOR_MAX_BYTES 16

#define SENSOR_P_DATA_BYTES 15
#define SENSOR_T_DATA_BYTES 16


/*---------------------------------------------*/
/* System Variables */
/*---------------------------------------------*/

static __IO uint32_t TimingDelay;
static __IO uint32_t TimeOut;


/*---------------------------------------------*/
/* Function Prototypes */
/*---------------------------------------------*/

int main(void);

void EXTI0_IRQHandler(void);
void USART2_IRQHandler(void);

void EnableClocks(void);
void ConfigureEXTI(void);
void ConfigureNVIC(void);
void ConfigureGPIO(void);
void ConfigureUSART2(void);
void ConfigureSPI2(void);
void ConfigureSPI3(uint16_t SPI_BaudRatePrescaler);
void ConfigureI2C2(void);
int ConfigureRTC(void);
int ConfigureSD(void);
int ConfigureSensors(void);
int ConfigureBluetooth(void);

void PinSet(GPIO_TypeDef* GPIO_PORT, uint16_t GPIO_PIN);
void PinReset(GPIO_TypeDef* GPIO_PORT, uint16_t GPIO_PIN);
uint16_t PortIDR(GPIO_TypeDef* GPIO_PORT);

int SendChar_USART2(uint16_t data);
int SendChar_SPI2(uint16_t data);
int SendChar_SPI3(uint16_t data);
int SendChar_I2C2(uint8_t reg_addr, uint8_t reg_data);
uint16_t ReceiveChar_USART2(void);
uint16_t ReceiveChar_SPI2(void);
uint16_t ReceiveChar_SPI3(void);
uint8_t ReceiveChar_I2C2(uint8_t reg_data);

void SendString_USART2(uint8_t data[]);
void SendChars_USART2(uint8_t data[], int numChars);
int ReceiveString_USART2(void);
int ReceiveChars_USART2(int numChars);

void I2C_RTC_Buffer(void);
uint8_t I2C_RTC_GetSeconds(void);
uint8_t I2C_RTC_GetMinutes(void);
uint8_t I2C_RTC_GetHours(void);
uint8_t I2C_RTC_GetDay(void);
uint8_t I2C_RTC_GetDate(void);
uint8_t I2C_RTC_GetMonth(void);
uint8_t I2C_RTC_GetYear(void);
void I2C_RTC_SetSeconds(uint8_t seconds);
void I2C_RTC_SetMinutes(uint8_t seconds);
void I2C_RTC_SetHours(uint8_t seconds);
void I2C_RTC_SetDay(uint8_t seconds);
void I2C_RTC_SetDate(uint8_t seconds);
void I2C_RTC_SetMonth(uint8_t seconds);
void I2C_RTC_SetYear(uint8_t seconds);
void I2C_RTC_Start(void);
void I2C_RTC_Stop(void);
void I2C_RTC_ResetDayDateTime(void);
void RTC_Set(char* dateString);

void CheckCmd(char* cmdStr);
void BT_Check(void);

void SEN_Select(int sensor);
void SEN_Release(void);
void SEN_GetData(void);
void SEN_FormatData(void);

void SD_CS_High(void);
void SD_CS_Low(void);
void FRESULTtoString(FRESULT fr);

int SD_Mount(void);
void SD_Unmount(void);
int SD_CreateLog(void);
int SD_OpenConfig(void);
int SD_CloseConfig(void);
int SD_CheckConfig(void);
int SD_OpenLog(void);
int SD_CloseLog(void);
int SD_LogReadLine(void);
int SD_LogWriteLine(void);

void Logging_Start(void);
void Logging_Stop(void);

void ClearBuffer_USART2(void);

#ifdef DEBUG_USE_USART1
	void USART1_IRQHandler(void);
	void ConfigureUSART1(void);
	int SendChar_USART1(uint16_t data);
	uint16_t ReceiveChar_USART1(void);
	void SendString_USART1(uint8_t data[]);
	void SendChars_USART1(uint8_t data[], int numChars);
	int ReceiveString_USART1(void);
	int ReceiveChars_USART1(int numChars);
	void ClearBuffer_USART1(void);
#endif

void Check_SD(void);

/*-----------------------------------------------------*/
/* Code below adapted from STMicroelectronics examples */
/*-----------------------------------------------------*/

//void TimingDelay_Decrement(void);
void Decrement_Counters(void);
void Delay(__IO uint32_t nTime);

#endif /* __MAIN_H */

/* END OF FILE */
